/* 
 * $Id: sysctl_impl.c,v 1.4 2002/01/20 22:20:00 mulix Exp $
 */

/* get the module's sysctl defines */
#include "sct_rules_module/module_interface.h"

#include "sct_ctrl_lib.h"
#include "sysctl_impl.h"

#include <malloc.h>
#include <assert.h>
#include <stdio.h>
#include <string.h>

#include <linux/unistd.h>
#include <linux/sysctl.h>

_syscall1(int, _sysctl, struct __sysctl_args *, args);

/* returns 0 on error and 1 on success */
static int do_sysctl(sct_ctrl_t ctrl, unsigned char* buf, 
		     unsigned int buflen, int* sysctl_cmd, 
		     unsigned int sysctl_cmd_len)
{
    struct __sysctl_args args =
	{ sysctl_cmd, sysctl_cmd_len, NULL, 0, buf, buflen};
    
    int rc;
    
    assert (sysctl_cmd);
    
    if (!sysctl_cmd)
	return 0;
    
    rc = _sysctl(&args);
    if (rc != 0) {
        fprintf(stderr, "sysctl returned '%d'\n", rc);
	perror("sysctl");
    }
    
    
    return (rc != -1) ? 1 : 0;
}

static int
sysctl_common(sct_ctrl_t ctrl, unsigned int sysctl_command, 
	      int rule_id, sct_rule_when_t when, int syscall_id, 
	      sct_filter_node_t* filter, sct_action_t* act,
	      int enable_rule)
{
    /* we ignore enable_rule for now */
    
    int serial_len;
    unsigned char buf[1000];
    int sysctl_cmd_struct[] = { CTL_TRACKER, sysctl_command };
    
    assert(rule_id >= 0 && syscall_id >= 0);
    
    if ((rule_id < 0) ||(syscall_id < 0))
	return 0;
    
    serial_len = 
	sct_serialize_tracking_rule(rule_id, when, syscall_id, act, 
				filter, buf, sizeof(buf));
    
    if (serial_len <= 0){
	fprintf(stderr, "failed to serialize tracking rule, "
		"error '%d'\n", serial_len);
	return 0;
    }

    return do_sysctl(ctrl, buf, serial_len, sysctl_cmd_struct, 
		     sizeof(sysctl_cmd_struct)/ sizeof(sysctl_cmd_struct[0]));
}

		     

/* add rule */
int 
sysctl_add_rule(sct_ctrl_t ctrl, int rule_id, sct_rule_when_t when, int syscall_id, 
		sct_filter_node_t* filter, sct_action_t* act,
		int enable_rule)
{
    assert (filter && act); 
    if (!filter || !act)
	return 0;

    return sysctl_common(ctrl, CTL_TRACKER_CMD_ADD_RULE, 
			 rule_id, when, syscall_id, 
			 filter, act, enable_rule);
}

/* delete a rule */
int
sysctl_del_rule(sct_ctrl_t ctrl, int rule_id, sct_rule_when_t when, int syscall_id)
{
    return sysctl_common(ctrl, CTL_TRACKER_CMD_DEL_RULE, 
			 rule_id, when, syscall_id, 
			 /* filter node */NULL, 
			 /* action */NULL, 0);
}

/* enable a rule */
int 
sysctl_enable_rule(sct_ctrl_t ctrl,
		   int rule_id,
		   sct_rule_when_t when,
		   int syscall_id)
{
    fprintf(stderr, "eek! sct_ctrl_enable_rule is unimplented!\n");
    return 0;
}

/* disable a rule */
int 
sysctl_disable_rule(sct_ctrl_t ctrl,
		    int rule_id,
		    sct_rule_when_t when,
		    int syscall_id)
{
    fprintf(stderr, "eek! sct_ctrl_disable_rule is unimplented!\n");
    return 0;
}

/* delete all rules for this syscall */
int 
sysctl_del_all_syscall_rules(sct_ctrl_t ctrl, int syscall_id)
{
    return sysctl_common(ctrl, CTL_TRACKER_CMD_DEL_ALL_RULES_FOR_SYSCALL, 
			 /* rule_id */ 0, /* when */ RW_BEFORE, syscall_id, 
			 /* filter node */NULL, 
			 /* action */NULL, 0);
}

/* delete all rules for ALL syscalls */
int 
sysctl_del_all_rules(sct_ctrl_t ctrl)
{
    return sysctl_common(ctrl, CTL_TRACKER_CMD_DEL_ALL_MODULE_RULES, 
			 /* rule_id */ 0, /* when */ RW_BEFORE,
			 /* syscall_id */ 0,
			 /* filter node */NULL, 
			 /* action */NULL, 0);
}

/* print the current rules */
int
sysctl_print_all_rules(sct_ctrl_t ctrl)
{
    return sysctl_common(ctrl, CTL_TRACKER_CMD_PRINT_ALL_RULES, 
			 /* rule_id */ 0, /* when */ RW_BEFORE,
			 /* syscall_id */ 0,
			 /* filter node */NULL, 
			 /* action */NULL, 0);
}
