#!/usr/bin/perl -w

use strict;

#
# $Id: gen_syscalls.pl,v 1.30 2002/02/21 03:07:10 chooga Exp $
#

# input/output files.

# list of struct types and syscalls to support.
my $data_file = "syscalls.dat";

# generated files
my $hijack_stubs_file = "syscall_hijack_autogen.c";
my $tracker_stubs_file = "syscall_tracker_autogen.c";

# template files used in the generation of the output files.
my $templates_dir = "./templates";
my $syscall_hijack_stub_tmpl =
	$templates_dir . '/' . "syscall_hijack_stub.tmpl";
my $syscall_tracker_stub_tmpl =
	$templates_dir . '/' . "syscall_tracker_stub.tmpl";
my $syscall_tracker_stub_after_tmpl =
	$templates_dir . '/' . "syscall_tracker_stub_after.tmpl";
my $syscall_tracker_mux_stub_tmpl =
	$templates_dir . '/' . "syscall_tracker_mux_stub.tmpl";
my $syscall_tracker_mux_stub_after_tmpl =
	$templates_dir . '/' . "syscall_tracker_mux_stub_after.tmpl";
my $syscall_tracker_struct_offsets_tmpl =
	$templates_dir . '/' . "syscall_tracker_struct_offsets.tmpl";
my $syscall_tracker_struct_log_tmpl =
	$templates_dir . '/' . "syscall_tracker_struct_log.tmpl";
my $syscall_tracker_struct_data_init_tmpl =
	$templates_dir . '/' . "syscall_tracker_struct_data_init.tmpl";

# data types mappings.
my %types_map = (
  'char', 'char',
  'uchar', 'unsigned char',
  'schar', 'signed char',
  'short', 'short',
  'ushort', 'unsigned short',
  'unsigned_short', 'unsigned short',
  'int', 'int',
  'uint', 'unsigned int',
  'unsigned_int', 'unsigned int',
  'long', 'long',
  'ulong', 'unsigned long',
  'unsigned_long', 'unsigned long',
  'string', 'char*',
  'const_string', 'const char*',
  'pointer', 'void*',
  'mode_t', 'mode_t',
  'size_t', 'size_t',
  'ssize_t', 'ssize_t',
  'pid_t', 'pid_t',
  'uid_t', 'uid_t',
  'void', 'void',
  'off_t', 'off_t',
  'net_ushort', 'unsigned short',
  'net_ip', 'unsigned long',
);

my %types_data_map = (
  'char', 'DT_CHAR',
  'short', 'DT_SHORT',
  'ushort', 'DT_USHORT',
  'unsigned_short', 'DT_USHORT',
  'int', 'DT_INT',
  'uint', 'DT_UINT',
  'unsigned_int', 'DT_UINT',
  'long', 'DT_LONG',
  'ulong', 'DT_ULONG',
  'unsigned_long', 'DT_ULONG',
  'string', 'DT_STRING',
  'const_string', 'DT_STRING',
  'mode_t', 'DT_USHORT',
  'size_t', 'DT_LONG',
  'ssize_t', 'DT_ULONG',
  'pid_t', 'DT_INT',
  'uid_t', 'DT_SHORT',
  'off_t', 'DT_LONG',
  'net_ushort', 'DT_USHORT',
  'net_ip', 'DT_ULONG',
);

my %types_makeval_map = (
  'char', 'sct_make_char_val',
  'short', 'sct_make_short_val',
  'ushort', 'sct_make_ushort_val',
  'unsigned_short', 'sct_make_ushort_val',
  'int', 'sct_make_int_val',
  'uint', 'sct_make_uint_val',
  'unsigned_int', 'sct_make_uint_val',
  'long', 'sct_make_long_val',
  'ulong', 'sct_make_ulong_val',
  'unsigned_long', 'sct_make_ulong_val',
  'string', 'sct_make_string_val',
  'const_string', 'sct_make_string_val',
  'mode_t', 'sct_make_ushort_val',
  'size_t', 'sct_make_long_val',
  'ssize_t', 'sct_make_ulong_val',
  'pid_t', 'sct_make_int_val',
  'uid_t', 'sct_make_ushort_val',
  'off_t', 'sct_make_long_val',
  'net_ushort', 'sct_make_ushort_val',
  'net_ip', 'sct_make_ulong_val',
);

my %types_specifier_map = (
  'char', 'c',
  'short', 'h',
  'ushort', 'uh',
  'unsigned_short', 'uh',
  'int', 'i',
  'uint', 'ui',
  'unsigned_int', 'ui',
  'long', 'l',
  'ulong', 'ul',
  'unsigned_long', 'ul',
  'string', 'P',
  'const_string', 'P',
  'mode_t', 'uh',
  'size_t', 'l',
  'ssize_t', 'ul',
  'pid_t', 'i',
  'uid_t', 'uh',
  'void', 'v',
  'off_t', 'l',
  'net_ushort', 'U',
  'net_ip', 'U',
);

my %struct_field_logger_helper_map = (
  'net_ushort', 'LHF_NTOHS',
  'net_ip', 'LHF_INET_NTOA',
);

my %basic_types_dummy_data_specifier_map = (
  'c', "'\\0'",
  'h', '0',
  'uh', '0',
  'i', '0',
  'ui', '0',
  'l', '0L',
  'ul', '0L',
  's', 'NULL',
  'v', 'NULL'
);

my ($matchable) = "";
my ($param_names_for_logging_string);
my ($param_struct_constness);
my ($syscall_param_matchable_ref);
my ($syscall_param_types_ref);
my ($syscall_param_names_ref);


my $os = `uname -s`;
chomp($os);

if ($#ARGV != 0) {
	die "Usage - $0 <syscalls number>";
}

my $num_syscalls = $ARGV[0];

my $cvs_id = "/*\n * \$Id\$\n */\n\n";

# read template files first.
open(TMPL, "<$syscall_hijack_stub_tmpl.$os") ||
	die "Cannot open file '$syscall_hijack_stub_tmpl.$os' for reading - $!";
my @syscall_hijack_stub_tmpl = <TMPL>;
close(TMPL);

open(TMPL, "<$syscall_tracker_stub_tmpl.$os") ||
	die "Cannot open file '$syscall_tracker_stub_tmpl.$os' for reading - $!";
my @syscall_tracker_stub_tmpl = <TMPL>;
close(TMPL);

open(TMPL, "<$syscall_tracker_stub_after_tmpl.$os") ||
	die "Cannot open file '$syscall_tracker_stub_after_tmpl' for reading - $!";
my @syscall_tracker_stub_after_tmpl = <TMPL>;
close(TMPL);

open(TMPL, "<$syscall_tracker_mux_stub_tmpl.$os") ||
        die "Cannot open file '$syscall_tracker_mux_stub_tmpl.$os' for reading - $!";
my @syscall_tracker_mux_stub_tmpl = <TMPL>;
close(TMPL);

open(TMPL, "<$syscall_tracker_mux_stub_after_tmpl.$os") ||
        die "Cannot open file '$syscall_tracker_mux_stub_after_tmpl.$os' for reading - $!";
my @syscall_tracker_mux_stub_after_tmpl = <TMPL>;
close(TMPL);

open(TMPL, "<$syscall_tracker_struct_offsets_tmpl.$os") ||
	die "Cannot open file '$syscall_tracker_struct_offsets_tmpl.$os' for reading - $!";
my @syscall_tracker_struct_offsets_tmpl = <TMPL>;
close(TMPL);

open(TMPL, "<$syscall_tracker_struct_log_tmpl.$os") ||
	die "Cannot open file '$syscall_tracker_struct_log_tmpl.$os' for reading - $!";
my @syscall_tracker_struct_log_tmpl = <TMPL>;
close(TMPL);

open(TMPL, "<$syscall_tracker_struct_data_init_tmpl.$os") ||
	die "Cannot open file '$syscall_tracker_struct_data_init_tmpl.$os' for reading - $!";
my @syscall_tracker_struct_data_init_tmpl = <TMPL>;
close(TMPL);


# open data file, scan and check input, generate output along the way.
open(DAT, "<$data_file") ||
	die "Cannot open file '$data_file' for reading - $!";
print "generating system calls stubs - $num_syscalls system calls\n";

if (-e $hijack_stubs_file) {
    rename("$hijack_stubs_file", "${hijack_stubs_file}.old") ||
	    die "Cannot rename '$hijack_stubs_file' to '${hijack_stubs_file}.old' - $!";
}
if (-e $tracker_stubs_file) {
    rename("$tracker_stubs_file", "${tracker_stubs_file}.old") ||
	    die "Cannot rename '$tracker_stubs_file' to '${tracker_stubs_file}.old' - $!";
}
open(HIJACK_AUTOGEN, ">$hijack_stubs_file") ||
	die "Cannot open '$hijack_stubs_file' for writing - $!";
open(TRACKER_AUTOGEN, ">$tracker_stubs_file") ||
	die "Cannot open '$tracker_stubs_file' for writing - $!";

print HIJACK_AUTOGEN $cvs_id;
print HIJACK_AUTOGEN "/* This file is automatically generated - do not edit. */\n";
print HIJACK_AUTOGEN "/* built by '$0 $num_syscalls' for $os. */\n\n";

if ($os eq "Linux") {
    print HIJACK_AUTOGEN "\n#include <asm/errno.h>\n\n";
} elsif ($os eq "FreeBSD") {
    print HIJACK_AUTOGEN "\n#include <sys/types.h>\n#include <sys/errno.h>\n#include <sys/sysproto.h>\n\n";
}

print TRACKER_AUTOGEN $cvs_id;
print TRACKER_AUTOGEN "/* This file is automatically generated - do not edit. */\n";
print TRACKER_AUTOGEN "/* built by '$0 $num_syscalls' for $os. */\n\n";
if ($os eq "Linux") {
    print TRACKER_AUTOGEN "\n#include <asm/errno.h>\n\n";
    print TRACKER_AUTOGEN "#include \"var_callbacks.h\"\n";
    print TRACKER_AUTOGEN "#include \"struct_handling.h\"\n";
    print TRACKER_AUTOGEN "\n";
} elsif ($os eq "FreeBSD") {
    print TRACKER_AUTOGEN "\n#include <sys/types.h>\n#include <sys/errno.h>\n#include <sys/sysproto.h>\n\n";
}

my @syscall_ids = ();
my @syscall_functions = ();
my %structs_fields = ();
my %structs_types_indexes = ();
my %syscall_track_mux_init = ();

my ($include_file);
my ($asis);
my ($cpp);
my ($struct_type, $struct_fields, $kernel_function);
my ($syscall_id, $syscall_name, $syscall_params);

my ($first_struct_function) = 'yes';
my ($first_stub_function) = 'yes';
my ($struct_type_index) = 0;

while (<DAT>) {
	chomp;
	s/#.*$//;               # strip comments.
	/^\s*$/ && next;	# skip empty lines.
	# handle 'include' lines:
	if (/^\s*include\s*:\s*([^\s]+)\s*$/) {
		$include_file = $1;
		print HIJACK_AUTOGEN "#include $include_file\n";
		next;
	}
	# handle 'asis' lines:
	if (/^\s*asis:\s*(.*)$/) {
		$asis = $1;
		print HIJACK_AUTOGEN "$asis", "\n";
		next;
	}
	# handle 'cpp' lines:
	if (/^\s*cpp:\s*(.*)$/) {
		$cpp = $1;
		print HIJACK_AUTOGEN "#${cpp}", "\n";
		next;
	}
	# handle 'struct' lines:
	if (/^\s*struct\s+([^\(\s]+)\s*\(([^\)]+)\)\s*;\s*$/) {
		$struct_type = $1;
		$struct_fields = $2;
		print "got struct '$struct_type', fields '$struct_fields'\n";
		$structs_fields{$struct_type} = $struct_fields;
		$structs_types_indexes{$struct_type} = $struct_type_index;
		$struct_type_index++;
		if ($first_struct_function eq 'yes') {
			print TRACKER_AUTOGEN "/*\n";
			print TRACKER_AUTOGEN " * struct-offsets evaluation functions for dynamic type-casting.\n";
			print TRACKER_AUTOGEN " */\n\n";
			$first_struct_function = 'no';
	 	}
		&emit_struct_function($data_file, $., $struct_type);
		next;
	}
	# syscall format:
	#kernel_function: syscall_ID: syscall_name(param_type param_name...);
	if (/^\s*([^:\s]+)\s*:\s*([^:\s]+)\s*:\s*([^\(\s]+)\s*\(([^\)]+)\)\s*;\s*$/) {
		$kernel_function = $1;
		$syscall_id = $2;
		$syscall_name = $3;
		$syscall_params = $4;
		push(@syscall_ids, $syscall_id);
		push(@syscall_functions, $syscall_name);
		print "kernel_function - '$kernel_function'\n";
		print "    syscall_id - '$syscall_id'\n";
		print "    syscall_name - '$syscall_name'\n";
		#print "    syscall_params - '$syscall_params'\n";

    		if ($first_stub_function eq 'yes') {
			print HIJACK_AUTOGEN "/*\n";
			print HIJACK_AUTOGEN " * Stub functions for syscall hijacking.\n";
			print HIJACK_AUTOGEN " */\n\n";
			print TRACKER_AUTOGEN "/*\n";
			print TRACKER_AUTOGEN " * Tracking functions for syscall hijacking.\n";
			print TRACKER_AUTOGEN " */\n\n";
			$first_stub_function = 'no';
    		}

		&emit_tracker_and_hijack_stubs($data_file,
					       $.,
					       $kernel_function,
					       $syscall_id,
					       $syscall_name,
					       $syscall_params, 
					       \%syscall_track_mux_init);
		next;
	}
	else {
		die "Invalid syntax in '$data_file' line $.";
	}
}

print TRACKER_AUTOGEN "\n";
print TRACKER_AUTOGEN "void\n";
print TRACKER_AUTOGEN "init_mux_functions(void)\n";
print TRACKER_AUTOGEN "{\n";
foreach $syscall_id (keys %syscall_track_mux_init) {
    print TRACKER_AUTOGEN "$syscall_track_mux_init{$syscall_id}";
    print TRACKER_AUTOGEN "\n";
}
print TRACKER_AUTOGEN "}\n";

print HIJACK_AUTOGEN "\n";
print TRACKER_AUTOGEN "\n";

&emit_structs_data_array(\%structs_fields, \%structs_types_indexes);
&emit_stubs_arrays($num_syscalls, \@syscall_ids);

close(HIJACK_AUTOGEN) ||
	die "Error closing '$hijack_stubs_file' - $!";
close(TRACKER_AUTOGEN) ||
	die "Error closing '$tracker_stubs_file' - $!";

exit(0);

# emit functions that allow calculating offsets of fields in the given struct.
sub emit_struct_function {
    my($data_file, $data_file_line, $struct_type) = @_;
    my($struct_params) = $structs_fields{$struct_type};
    my(@struct_params_strings) = split(',', $struct_params);
    my(@pre_struct_param_matchable) = ();
    my( @struct_param_matchable) = ();
    my(@pre_struct_param_types) = ();
    my( @struct_param_types) = ();
    my(@pre_struct_param_names) = ();
    my( @struct_param_names) = ();
    my(@pre_struct_param_isptr) = ();
    my( @struct_param_isptr) = ();
    my(@pre_struct_param_isarray) = ();
    my( @struct_param_isarray) = ();
    my(@pre_struct_param_array_len) = ();
    my( @struct_param_array_len) = ();
    my(@pre_struct_param_fmt_spec) = ();
    my( @struct_param_fmt_spec) = ();
    my($line, $line_copy);
    my($pre_num_struct_fields, $num_struct_fields);
    my($field_offsets_calc) = "";
    my($field_types_calc) = "";
    my($i);

    &break_down_params_list($data_file, $data_file_line,
			    \@struct_params_strings,
			    \@pre_struct_param_matchable,
			    \@pre_struct_param_types,
			    \@pre_struct_param_names,
			    \@pre_struct_param_isptr,
			    \@pre_struct_param_isarray,
			    \@pre_struct_param_array_len,
			    \@pre_struct_param_fmt_spec);

    # handle struct-in-struct
    &expand_struct_in_struct($data_file, $data_file_line,
			    \@pre_struct_param_matchable,
			    \@pre_struct_param_types,
			    \@pre_struct_param_names,
			    \@pre_struct_param_isptr,
			    \@pre_struct_param_isarray,
			    \@pre_struct_param_array_len,
			    \@pre_struct_param_fmt_spec,
			    \@struct_param_matchable,
			    \@struct_param_types,
			    \@struct_param_names,
			    \@struct_param_isptr,
			    \@struct_param_isarray,
			    \@struct_param_array_len,
			    \@struct_param_fmt_spec);

    $num_struct_fields = $#struct_param_names + 1;

    # prepare code to calculate the offset+type for each field of the struct.
    for ($i=0; $i < $num_struct_fields; $i++) {
	if ($struct_param_matchable[$i] eq 'matchable') {
	    $field_offsets_calc .= "\t\toffsets_vec[$i] = (long)( ((char*)&s.$struct_param_names[$i]) - base);\n";
	    $field_types_calc .= "\t\ttypes_vec[$i] = $types_data_map{$struct_param_types[$i]};\n";
	}
	else {
	    $field_offsets_calc .= "\t\toffsets_vec[$i] = -1;\n";
	    $field_types_calc .= "\t\ttypes_vec[$i] = DT_LAST_TYPE;\n";
	}
    }

    my (@struct_param_names_deref) = ();
    for ($i=0; $i < $num_struct_fields; $i++) {
	if (defined($struct_field_logger_helper_map{$struct_param_types[$i]})) {
	    my $lhf_func =
		$struct_field_logger_helper_map{$struct_param_types[$i]};
	    $struct_param_names_deref[$i] =
		'&s->' . $struct_param_names[$i] . ", (void*)$lhf_func";
	}
	else {
	    $struct_param_names_deref[$i] = 's->' . $struct_param_names[$i];
	}
    }
    my($struct_fields_names) = join(', ', @struct_param_names_deref);
    my($struct_param_fmt_spec_str) = join(' ', @struct_param_fmt_spec);

    # generate the functions using a template file.
    foreach $line (@syscall_tracker_struct_offsets_tmpl) {
	# avoid modifying the syscall_hijack_stub_tmpl array - thus,
	# work on a copy of that line.
	$line_copy = $line;
	$line_copy =~ s/%%STRUCT_TYPE%%/$struct_type/g;
	$line_copy =~ s/%%STRUCT_TYPE_NUM_FIELDS%%/$num_struct_fields/g;
	$line_copy =~ s/%%STRUCT_FIELDS_OFFSET_CALC%%/$field_offsets_calc/g;
	$line_copy =~ s/%%STRUCT_FIELDS_TYPES_CALC%%/$field_types_calc/g;
        print TRACKER_AUTOGEN "$line_copy";
    }

    foreach $line (@syscall_tracker_struct_log_tmpl) {
	# avoid modifying the syscall_tracker_struct_log_tmpl array - thus,
	# work on a copy of that line.
	$line_copy = $line;
	$line_copy =~ s/%%STRUCT_TYPE%%/$struct_type/g;
	$line_copy =~ s/%%STRUCT_TYPE_NUM_FIELDS%%/$num_struct_fields/g;
	$line_copy =~ s/%%STRUCT_FIELDS_NAMES%%/$struct_fields_names/g;
	$line_copy =~ s/%%STRUCT_LOG_SPEC%%/$struct_param_fmt_spec_str/g;
        print TRACKER_AUTOGEN "$line_copy";
    }

    print TRACKER_AUTOGEN "\n";
}

sub emit_tracker_and_hijack_stubs {
    my($data_file, $data_file_line,
       $kernel_function, $syscall_id, $syscall_name,
       $syscall_params, $syscall_track_mux_init_ref) = @_;
    my($syscall_params_strings_ref ) = @_;
    my(@syscall_param_matchable) = ();
    my(@syscall_param_types) = ();
    my(@syscall_param_names) = ();
    my(@syscall_param_isptr) = ();
    my(@syscall_param_isarray) = ();
    my(@syscall_param_array_len) = ();
    my(@syscall_param_fmt_spec) = ();
    my($params_logging_spec);
    my(@syscall_params_strings) = split(',', $syscall_params);
    my($param_matchable, $param_type, $param_name, $param_isptr);
    my($param_isarray, $param_array_len);
    my($param_makeval);
    my($param_datatype);
    my($struct_param_matchable, $struct_param_type, $struct_param_name);
    my($struct_param_makeval, $struct_logging_params);
    my($i, $j);
    my($param_output_type, $param_output_string, $param_names_string);
    my($line, $line_copy);
    my($syscall_copied_vars_def, $param_struct_constness, $param_struct_type);
    my($syscall_copied_vars_copy, $syscall_copied_vars_goto_free);
    my($syscall_params_filter_make, $syscall_params_struct_filter_assign);
    my($syscall_params_filter_free);
    my($syscall_params_filter_vars_set, $syscall_params_filter_retval_var_set);
    my($syscall_copied_vars_free, $first_putstring_function);
    my($is_void_syscall) = 0;
    my($is_mux_func, $mux_func_id) = 0;
    my($before_syscall_params_recv, $after_syscall_params_recv);
    my($after_retval_func_param);
    my($after_retval_sys_ret);
    my($before_retval_variable);


    # check if this 'syscall id' actually belongs to a muxed function.
    $mux_func_id = ($syscall_id >> 16) & 0xffff;
    $syscall_id = $syscall_id & 0xffff;
    if ($mux_func_id > 0) {
	$is_mux_func = 1;
	print "    syscall $syscall_id - muxes function $mux_func_id\n";
	if (!defined(%$syscall_track_mux_init_ref->{$syscall_id})) {
	    %$syscall_track_mux_init_ref->{$syscall_id} =
		"\t/* init for mux syscall '$syscall_id'. */\n";
	}
	%$syscall_track_mux_init_ref->{$syscall_id} .=
		"\tinit_mux_function(${syscall_id}, ${mux_func_id}, (func_ptr)track_before_syscall_${syscall_id}_${mux_func_id}, (func_ptr)track_after_syscall_${syscall_id}_${mux_func_id});\n";
    }

    # break down the parameters list.
    &break_down_params_list($data_file, $data_file_line,
			    \@syscall_params_strings,
			    \@syscall_param_matchable,
			    \@syscall_param_types,
			    \@syscall_param_names,
			    \@syscall_param_isptr,
			    \@syscall_param_isarray,
			    \@syscall_param_array_len,
			    \@syscall_param_fmt_spec);

    print "    params num: ", $#syscall_param_types + 1, "\n";
    for ($i = 0; $i <= $#syscall_param_types; $i++) {
      print "     param $i: ";
      print ""
	. (($syscall_param_matchable[$i] eq 'matchable') ? $matchable : "")
	. "type '$syscall_param_types[$i]'"
	. "  name '" . (defined($syscall_param_names[$i]) ?
		      $syscall_param_names[$i] : "")
	. "'\n";
    }

    # prepare code for param logging spec, including logging of
    # struct members.
    for ($i = 0; $i <= $#syscall_param_types; $i++) {
	$param_type = $syscall_param_types[$i];
	$param_name = $syscall_param_names[$i];
	$param_matchable = $syscall_param_matchable[$i];
	# if its a struct - prepare a log spec for this struct.
	if ($param_type =~ /^(const_)?struct_(.*)$/) {
	    $struct_type = $2;
	    if ($param_matchable eq 'matchable') {
	        my($struct_params) = $structs_fields{$struct_type};
    	        my(@struct_params_strings) = split(',', $struct_params);
	        my(@pre_struct_param_matchable) = ();
	        my(@pre_struct_param_types) = ();
	        my(@pre_struct_param_names) = ();
	        my(@pre_struct_param_isptr) = ();
	        my(@pre_struct_param_isarray) = ();
	        my(@pre_struct_param_array_len) = ();
	        my(@pre_struct_param_fmt_spec) = ();
	        my(@struct_param_matchable) = ();
	        my(@struct_param_types) = ();
	        my(@struct_param_names) = ();
	        my(@struct_param_isptr) = ();
	        my(@struct_param_isarray) = ();
	        my(@struct_param_array_len) = ();
	        my(@struct_param_fmt_spec) = ();

    	        &break_down_params_list($data_file, $data_file_line,
			    	        \@struct_params_strings,
			    	        \@pre_struct_param_matchable,
			    	        \@pre_struct_param_types,
			    	        \@pre_struct_param_names,
			    	        \@pre_struct_param_isptr,
			    	        \@pre_struct_param_isarray,
			    	        \@pre_struct_param_array_len,
			    	        \@pre_struct_param_fmt_spec);

    		&expand_struct_in_struct($data_file, $data_file_line,
			    		\@pre_struct_param_matchable,
			    		\@pre_struct_param_types,
			    		\@pre_struct_param_names,
			    		\@pre_struct_param_isptr,
			    		\@pre_struct_param_isarray,
			    		\@pre_struct_param_array_len,
			    		\@pre_struct_param_fmt_spec,
			    		\@struct_param_matchable,
			    		\@struct_param_types,
			    		\@struct_param_names,
			    		\@struct_param_isptr,
			    		\@struct_param_isarray,
			    		\@struct_param_array_len,
			    		\@struct_param_fmt_spec);

		# not in use - struct params are now logged via callbacks,
		#   not directly.
	        #$syscall_param_fmt_spec[$i] =
		#    "S{" . join(' ', @struct_param_fmt_spec) . "}";
	        $syscall_param_fmt_spec[$i] = "S";
	    }
	    else {
	        $syscall_param_fmt_spec[$i] = "p";
	    }
	}
    }
    $params_logging_spec = join(' ', @syscall_param_fmt_spec);
    print "params_spec: '", $params_logging_spec, "'\n";

    # generate parameters definition string.
    if ($os eq "Linux") {
	$param_output_string = '';
	$param_names_string = '';
	$param_names_for_logging_string = '';
	for ($i = 0; $i <= $#syscall_param_types; $i++) {
	    $param_type = $syscall_param_types[$i];
	    $param_name = $syscall_param_names[$i];
	    $param_isptr = $syscall_param_isptr[$i];
	    $param_isarray = $syscall_param_isarray[$i];
	    $param_array_len = $syscall_param_array_len[$i];
	    $param_output_type = $types_map{$param_type};
	    # void handling
	    if ($param_type eq "void") {
		$is_void_syscall = 1;
	    }
	    # special handling for 'struct' types.
	    if ($param_type =~ /^(const_)?struct_(.*)$/) {
		$param_struct_constness = $1;
		$param_struct_type = $2;
		if (!defined($param_struct_constness)) {
		    $param_struct_constness = "";
		}
		$param_struct_constness =~ s/_/ /g; # remove trailing '_'.
		if (!defined($structs_fields{$param_struct_type})) {
		    die "Invalid parameter struct type '$param_type' "
		        . "in '$data_file' line $data_file_line";
		}
		$param_output_type =
		    $param_struct_constness . "struct " . $param_struct_type . "*";
	    }
	    if ((!defined($param_output_type) || ($param_output_type eq '')) &&
			($is_void_syscall != 1)) {
		die "Invalid parameter type '$param_type' "
		    . "in '$data_file' line $data_file_line";
	    }
	    $param_output_string .= $param_output_type;
	    if ($param_isptr == 1) {
	        $param_output_string .= '*';
	    }
	    $param_output_string .= defined($param_name) ? " $param_name" : "";
	    if ($param_isarray) {
		$param_output_string .= '[';
		if ($param_array_len >= 1) {
		    $param_output_string .= $param_array_len;
		}
		$param_output_string .= ']';
	    }

	    $param_output_string .=  ", ";
	    
	    if (defined($types_specifier_map{$param_type}) ||
		$param_type =~ /^(const_)?struct_(.*)$/) {
	        if ($param_type =~ /^(const_)?struct_(.*)$/) {
		    $param_names_for_logging_string .= "";
		    # handle passing the members of the struct itself.
	    	    $struct_type = $2;
	    	    my($struct_params) = $structs_fields{$struct_type};
    	    	    my(@struct_params_strings) = split(',', $struct_params);
    	    	    $struct_logging_params =
			&prepare_struct_logging_params($data_file,
						       $data_file_line,
						       $param_name,
			    	    		       \@struct_params_strings);
	        }
                $param_names_string .= (defined($param_name) ? "$param_name" : "");
	        if (defined($param_type) &&
			$param_type =~ /^(const_)?struct_(.*)$/) {
	    	     $struct_type = $2;
		     my($struct_type_index) = $structs_types_indexes{$struct_type};
		     $param_names_for_logging_string .=
			"${param_name}, (void*)&${param_name}_logger_data";
		# not in use - struct params are now logged via callbacks,
                #  not directly.
		#    $param_names_for_logging_string .= 
		#	"\"$param_struct_type\", ${param_name}";
		#    if ($param_matchable eq 'matchable') {
		#	$param_names_for_logging_string .=
		#		", " . $struct_logging_params;
		#    }
		}
	        elsif ($param_type =~ /^string$|^const_string$/) {
		    $param_names_for_logging_string .= "$i";
		} 
		else {
		    $param_names_for_logging_string .= (defined($param_name) ? "$param_name" : "");
		}
		if ($i < $#syscall_param_types) {
		    $param_names_string .= ", ";
		    $param_names_for_logging_string .= ", ";
		}
	    }
	}
    } elsif ($os eq "FreeBSD") {
	$param_output_string =
	    "struct proc *proc, struct ${syscall_name}_args *uap";
	$param_names_string = "proc, uap";
	$param_names_for_logging_string = "proc, uap";
    }

    # prepare code for defining variables for copying strings and structs
    # from user-space.
    $syscall_copied_vars_def = "";
    for ($i = 0; $i <= $#syscall_param_types; $i++) {
	$param_type = $syscall_param_types[$i];
	$param_name = $syscall_param_names[$i];
	$param_matchable = $syscall_param_matchable[$i];
	if ($param_matchable eq 'matchable' && 
	    $param_type =~ /^(const_)?struct_(.*)$/) {
		$param_struct_type = $2;
	        $syscall_copied_vars_def .=
		    "\tstruct $param_struct_type kern_${param_name};\n" .
		    "\tstruct {\n" .
		    "\t    char* data;\n" .
		    "\t    int struct_type;\n" .
		    "\t} ${param_name}_struct_data;\n" .
		    "\tstruct struct_logger_data ${param_name}_logger_data;\n";
	}
    }

    # prepare code for copying strings and structs from user-space.
    $syscall_copied_vars_copy = "";

    # default action to take if we hit an error is to just return, 
    # since we might do not necessarily have a a label to 'goto'.
    $syscall_copied_vars_goto_free = "\t\t\treturn retval;\n";
    for ($i = 0; $i <= $#syscall_param_types; $i++) {
	$param_type = $syscall_param_types[$i];
	$param_name = $syscall_param_names[$i];
	$param_matchable = $syscall_param_matchable[$i];
	if ($param_matchable eq 'matchable' &&
	    $param_type =~ /^(const_)?struct_(.*)$/) {
		$param_struct_constness = $1;
		$param_struct_type = $2;
		if (!defined($param_struct_constness)) {
		    $param_struct_constness = "";
		}
		$param_struct_constness =~ s/_/ /g; # remove trailing '_'.
	        $syscall_copied_vars_copy .= "\tif (${param_name}) {\n";
		if ($os eq "Linux") {
	            $syscall_copied_vars_copy .= "\t\tif (copy_from_user(&kern_${param_name}, ${param_name}, sizeof(kern_${param_name}))) {\n";
		    $syscall_copied_vars_copy .= "\t\t\tretval = -EFAULT;\n";
		    $syscall_copied_vars_copy .= "\t\t\tgoto tracker_${syscall_id}_failed;\n";
		    $syscall_copied_vars_copy .= "\t\t}\n";
		    $syscall_copied_vars_copy .= "\t\t${param_name}_struct_data.data = (char*)&kern_${param_name};\n";
		    $syscall_copied_vars_copy .= "\t\t${param_name}_struct_data.struct_type = $structs_types_indexes{$param_struct_type};\n";
		    $syscall_copied_vars_copy .= "\t\t${param_name}_logger_data.param_index = ${i};\n";
		    $syscall_copied_vars_copy .= "\t\t${param_name}_logger_data.struct_type = $structs_types_indexes{$param_struct_type};\n";
		} elsif ($os eq "FreeBSD") {
		    # TODO: handle FreeBSD
		}
	        $syscall_copied_vars_copy .= "\t}\n";

		# make sure to generate code to free them later if we run into errors
		$syscall_copied_vars_goto_free = "\t\t\tgoto tracker_${syscall_id}_failed;\n";
	}
    }

# the following code isn't used any more, since we do NOT assign any
# parameter types before they are actualy needed any more - only via
# variable callbacks.

#    # prepare code for generating function parameters variables, and freeing them.
    $syscall_params_filter_make = "";
    $syscall_params_filter_free = "";
#    for ($i = 0; $i <= $#syscall_param_types; $i++) {
#	$param_type = $syscall_param_types[$i];
#	$param_name = $syscall_param_names[$i];
#	$param_matchable = $syscall_param_matchable[$i];
#	$param_makeval = $types_makeval_map{$param_type};
#	if ($os eq "FreeBSD") {
#	    $param_name = "uap->${param_name}";
#	}
#	# special handling for 'struct' types.
#	if ($param_matchable eq 'matchable' && $param_type =~ /^(const_)?struct_(.*)$/) {
#	    $struct_type = $2;
#	    my($struct_params) = $structs_fields{$struct_type};
#    	    my(@struct_params_strings) = split(',', $struct_params);
#	    my(@struct_param_matchable) = ();
#	    my(@struct_param_types) = ();
#	    my(@struct_param_names) = ();
#	    my(@struct_param_isptr) = ();
#	    my(@struct_param_isarray) = ();
#	    my(@struct_param_array_len) = ();
#	    my(@struct_param_fmt_spec) = ();
#    	    &break_down_params_list($data_file, $data_file_line,
#			    	    \@struct_params_strings,
#			    	    \@struct_param_matchable,
#			    	    \@struct_param_types,
#			    	    \@struct_param_names,
#			    	    \@struct_param_isptr,
#			    	    \@struct_param_isarray,
#			    	    \@struct_param_array_len,
#			    	    \@struct_param_fmt_spec);
#	    my($struct_params_num) = $#struct_param_types + 1;
#	    $syscall_params_filter_make .= "\t\tsct_filter_node_t* param${i} = sct_make_filter_vector(${struct_params_num});\n";
#	    $syscall_params_filter_free .= "\t\t\tsct_free_filter_node(param${i});\n";
#	}
#    }
#
# End of commented-out code.

# the following code isn't used any more, since we do NOT assign any
# parameter types before they are actualy needed any more - only via
# variable callbacks.

#    # prepare code for filling function struct parameters variables.
    $syscall_params_struct_filter_assign = "";
#    for ($i = 0; $i <= $#syscall_param_types; $i++) {
#	$param_type = $syscall_param_types[$i];
#	$param_name = $syscall_param_names[$i];
#	$param_matchable = $syscall_param_matchable[$i];
#	if ($param_matchable eq 'matchable' && $param_type =~ /^(const_)?struct_(.*)$/) {
#	    my($struct_type) = $2;
#	    my($struct_params) = $structs_fields{$struct_type};
#    	    my(@struct_params_strings) = split(',', $struct_params);
#	    my(@struct_param_matchable) = ();
#	    my(@struct_param_types) = ();
#	    my(@struct_param_names) = ();
#	    my(@struct_param_isptr) = ();
#	    my(@struct_param_isarray) = ();
#	    my(@struct_param_array_len) = ();
#	    my(@struct_param_fmt_spec) = ();
#    	    &break_down_params_list($data_file, $data_file_line,
#			    	    \@struct_params_strings,
#			    	    \@struct_param_matchable,
#			    	    \@struct_param_types,
#			    	    \@struct_param_names,
#			    	    \@struct_param_isptr,
#			    	    \@struct_param_isarray,
#			    	    \@struct_param_array_len,
#			    	    \@struct_param_fmt_spec);
#	    my($struct_params_num) = $#struct_param_types + 1;
#	    $syscall_params_struct_filter_assign .= "\t\tif ($param_name) {\n";
#            for ($j = 0; $j <= $#struct_param_types; $j++) {
#		$struct_param_type = $struct_param_types[$j];
#		$struct_param_name = $struct_param_names[$j];
#		$struct_param_matchable = $struct_param_matchable[$j];
#	        $struct_param_makeval = $types_makeval_map{$struct_param_type};
#	        if ($struct_param_type =~ /^string$|^const_string$/) {
#	            $struct_param_name = "kern_${param_name}";
#	        } elsif ($os eq "FreeBSD") {
#	            $struct_param_name = "uap->${param_name}";
#	        }
#		print "PARAM: ${i}.${j} - $struct_param_type, "
#			. "$struct_param_name, $struct_param_matchable, ";
#		print (defined(${struct_param_makeval}) ? $struct_param_makeval : "");
#		print "\n";
#		if ($struct_param_matchable eq 'matchable' && defined($struct_param_makeval) && $struct_param_makeval ne '') {
#	    	    $syscall_params_struct_filter_assign .= "\t\t\tsct_insert_filter_node(param${i}, $j, ${struct_param_makeval}(kern_${param_name}.${struct_param_name}));\n";
#		}
#	    }
#	    $syscall_params_struct_filter_assign .= "\t\t}\n";
#	}
#    }
#
# End of commented-out code.

    # prepare code for generating function parameters variable setting calls.
    $syscall_params_filter_vars_set = "";
    for ($i = 0; $i <= $#syscall_param_types; $i++) {
	$param_type = $syscall_param_types[$i];
	$param_name = $syscall_param_names[$i];
	$param_matchable = $syscall_param_matchable[$i];
	$param_makeval = $types_makeval_map{$param_type};
	$param_datatype = $types_data_map{$param_type};
	if ($param_matchable eq 'matchable' && $param_type =~ /^(const_)?struct_(.*)$/) {
	    $struct_type = $2;
	    $syscall_params_filter_vars_set .= "\t\tset_var_indexed_field_callback(VT_PARAM, ${i}, global_struct_field_param_callback, DT_INT, (void*)&${param_name}_struct_data);\n";
	}
	elsif ($param_matchable eq 'matchable' && $param_type =~ /^string$|^const_string$/) {
	    $syscall_params_filter_vars_set .= "\t\tset_var_indexed_callback(VT_PARAM, ${i}, global_param_callback, ${param_datatype}, (void*)${param_name});\n";
	}
	elsif ($param_matchable eq 'matchable' && defined($param_makeval) && $param_makeval ne '') {
	    $syscall_params_filter_vars_set .= "\t\tset_var_indexed_callback(VT_PARAM, ${i}, global_param_callback, ${param_datatype}, (void*)&${param_name});\n";
	}
    }

    # prepare code for generating syscall retval variable setting call.
    $syscall_params_filter_retval_var_set = "";
    $syscall_params_filter_retval_var_set .=
	"\t\tset_var_callback(VT_RETVAL, global_retval_callback, (void*)syscall_rc);\n";

    # prepare code for generating process state variable setting calls.
    $syscall_copied_vars_free = "";
    # if there were structs, insert the failer
    # label and retval checking code.
    for ($i = 0; $i <= $#syscall_param_types; $i++) {
        $param_type = $syscall_param_types[$i];
        $param_name = $syscall_param_names[$i];
        $param_matchable = $syscall_param_matchable[$i];
        if ($param_matchable eq 'matchable' &&
    		$param_type =~ /^struct_|^const_struct_/) {
    	    $syscall_copied_vars_free .= 
		        "  tracker_${syscall_id}_failed:\n";
	    last;
	}
    }
    if ($is_void_syscall == 1) {  # for void type 
        $param_names_string = "";
        $param_output_string = "";
	$after_retval_func_param = "int syscall_rc";
	my $before_retval_func_param = "int* syscall_rc";
	$before_retval_variable = "&fake_syscall_retval";
	$after_retval_sys_ret = "orig_syscall_retval";
	$after_syscall_params_recv = $after_retval_func_param;
	$before_syscall_params_recv = $before_retval_func_param;
    }
    else {
	$param_names_for_logging_string =
		", " . $param_names_for_logging_string;
	$after_retval_func_param = "int syscall_rc,";
	my $before_retval_func_param = ", int* syscall_rc"; 
	$before_retval_variable = ", &fake_syscall_retval";
	$after_retval_sys_ret = "orig_syscall_retval,";
	$after_syscall_params_recv =
		$after_retval_func_param . " " . $param_output_string;
	$before_syscall_params_recv =  $param_output_string . "int* syscall_rc"; 
	# get rid of the superflous ', '
	chop($after_syscall_params_recv); 
	chop($after_syscall_params_recv); 
    }

    # generate the syscall hijack stub function for this syscall,
    # unless its not a syscall, but rather a muxed function.
    if ($is_mux_func == 0) {
        my $hijack_params_recv; 
	if ($param_output_string eq '') {
	    $hijack_params_recv = "void";
	} else {
	    $hijack_params_recv = $param_output_string;
	    # get rid of the superflous ', '
	    chop($hijack_params_recv); chop($hijack_params_recv);
	}
        foreach $line (@syscall_hijack_stub_tmpl) {
	    # avoid modifying the syscall_hijack_stub_tmpl array - thus,
	    # work on a copy of that line.
	    $line_copy = $line;
	    $line_copy =~ s/%%SYSCALL_ID%%/$syscall_id/g;
	    $line_copy =~ s/%%SYSCALL_NAME%%/$syscall_name/g;
	    $line_copy =~ s/%%AFTER_RETVAL_FUNC_PARAM%%/$after_retval_func_param/g;
	    $line_copy =~ s/%%AFTER_SYSCALL_PARAMS_RECV%%/$after_syscall_params_recv/g;
	    $line_copy =~ s/%%BEFORE_RETVAL_VARIABLE%%/$before_retval_variable/g;
	    $line_copy =~ s/%%AFTER_RETVAL_SYS_RET%%/$after_retval_sys_ret/g;
	    $line_copy =~ s/%%SYSCALL_PARAMS_RECV%%/$hijack_params_recv/g;
	    $line_copy =~ s/%%BEFORE_SYSCALL_PARAMS_RECV%%/$before_syscall_params_recv/g;
	    $line_copy =~ s/%%SYSCALL_PARAMS_NAMES%%/$param_names_string/g;
            print HIJACK_AUTOGEN "$line_copy";
	}
        print HIJACK_AUTOGEN "\n";
    }
    my(@tracker_stub_tmpl);
    @tracker_stub_tmpl = (($is_mux_func == 0) ? 
       @syscall_tracker_stub_tmpl :@syscall_tracker_mux_stub_tmpl);
	
    # generate a track function for the tracker module.
    foreach $line (@tracker_stub_tmpl) {
	$line_copy = $line; # avoid modifying the syscall_tracker_stub_tmpl array
	$line_copy =~ s/%%MUX_FUNC_ID%%/$mux_func_id/g;
	$line_copy =~ s/%%SYSCALL_ID%%/$syscall_id/g;
	$line_copy =~ s/%%SYSCALL_NAME%%/$syscall_name/g;
	$line_copy =~ s/%%SYSCALL_PARAMS_RECV%%/$param_output_string/g;
	$line_copy =~ s/%%SYSCALL_PARAMS_NAMES%%/$param_names_string/g;
	$line_copy =~ s/%%SYSCALL_COPIED_VARS_DEF%%/$syscall_copied_vars_def/g && chomp $line_copy;
	$line_copy =~ s/%%SYSCALL_COPIED_VARS_COPY%%/$syscall_copied_vars_copy/g && chomp $line_copy;
	$line_copy =~ s/%%SYSCALL_COPIED_VARS_GOTO_FREE%%/$syscall_copied_vars_goto_free/g && chomp $line_copy;
	$line_copy =~ s/%%SYSCALL_PARAMS_FILTER_MAKE%%/$syscall_params_filter_make/g && chomp $line_copy;
	$line_copy =~ s/%%SYSCALL_PARAMS_FILTER_FREE%%/$syscall_params_filter_free/g && chomp $line_copy;
	$line_copy =~ s/%%SYSCALL_PARAMS_STRUCT_FILTER_ASSIGN%%/$syscall_params_struct_filter_assign/g && chomp $line_copy;
	$line_copy =~ s/%%SYSCALL_PARAMS_FILTER_VARS_SET%%/$syscall_params_filter_vars_set/g && chomp $line_copy;
	$line_copy =~ s/%%SYSCALL_PARAMS_FILTER_RETVAL_VAR_SET%%/$syscall_params_filter_retval_var_set/g && chomp $line_copy;
	$line_copy =~ s/%%PARAMS_LOGGING_SPEC%%/$params_logging_spec/g;
	$line_copy =~ s/%%PARAM_NAMES_FOR_LOGGING%%/$param_names_for_logging_string/g;
	$line_copy =~ s/%%SYSCALL_COPIED_VARS_FREE%%/$syscall_copied_vars_free/g && chomp $line_copy;
        print TRACKER_AUTOGEN "$line_copy";
    }
    print TRACKER_AUTOGEN "\n";

    my(@tracker_stub_after_tmpl);
    @tracker_stub_after_tmpl = (($is_mux_func == 0) ? 
       @syscall_tracker_stub_after_tmpl :@syscall_tracker_mux_stub_after_tmpl);
	
    # generate a track function for the tracker module.
    foreach $line (@tracker_stub_after_tmpl) {
	$line_copy = $line; # avoid modifying the syscall_tracker_stub_after_tmpl array
	$line_copy =~ s/%%MUX_FUNC_ID%%/$mux_func_id/g;
	$line_copy =~ s/%%SYSCALL_ID%%/$syscall_id/g;
	$line_copy =~ s/%%SYSCALL_NAME%%/$syscall_name/g;
	$line_copy =~ s/%%SYSCALL_PARAMS_RECV%%/$param_output_string/g;
	$line_copy =~ s/%%SYSCALL_PARAMS_NAMES%%/$param_names_string/g;
	$line_copy =~ s/%%SYSCALL_COPIED_VARS_DEF%%/$syscall_copied_vars_def/g && chomp $line_copy;
	$line_copy =~ s/%%SYSCALL_COPIED_VARS_COPY%%/$syscall_copied_vars_copy/g && chomp $line_copy;
	$line_copy =~ s/%%SYSCALL_COPIED_VARS_GOTO_FREE%%/$syscall_copied_vars_goto_free/g && chomp $line_copy;
	$line_copy =~ s/%%SYSCALL_PARAMS_FILTER_MAKE%%/$syscall_params_filter_make/g && chomp $line_copy;
	$line_copy =~ s/%%SYSCALL_PARAMS_FILTER_FREE%%/$syscall_params_filter_free/g && chomp $line_copy;
	$line_copy =~ s/%%SYSCALL_PARAMS_STRUCT_FILTER_ASSIGN%%/$syscall_params_struct_filter_assign/g && chomp $line_copy;
	$line_copy =~ s/%%SYSCALL_PARAMS_FILTER_VARS_SET%%/$syscall_params_filter_vars_set/g && chomp $line_copy;
	$line_copy =~ s/%%SYSCALL_PARAMS_FILTER_RETVAL_VAR_SET%%/$syscall_params_filter_retval_var_set/g && chomp $line_copy;
	$line_copy =~ s/%%AFTER_SYSCALL_PARAMS_RECV%%/$after_syscall_params_recv/g;
	$line_copy =~ s/%%BEFORE_SYSCALL_PARAMS_RECV%%/$before_syscall_params_recv/g;
	$line_copy =~ s/%%AFTER_RETVAL_SYS_RET%%/$after_retval_sys_ret/g;
	$line_copy =~ s/%%PARAMS_LOGGING_SPEC%%/$params_logging_spec/g;
	$line_copy =~ s/%%PARAM_NAMES_FOR_LOGGING%%/$param_names_for_logging_string/g;
	$line_copy =~ s/%%SYSCALL_COPIED_VARS_FREE%%/$syscall_copied_vars_free/g && chomp $line_copy;
        print TRACKER_AUTOGEN "$line_copy";
    }
    print TRACKER_AUTOGEN "\n";
}

sub break_down_params_list {
    my($data_file, $data_file_line,
	  $syscall_params_strings_ref,
	  $syscall_param_matchable_ref,
	  $syscall_param_types_ref,
	  $syscall_param_names_ref,
	  $syscall_param_isptr_ref,
	  $syscall_param_isarray_ref,
	  $syscall_param_array_len_ref,
	  $syscall_param_fmt_spec_ref) = @_;

    my($syscall_param_string);
    my($param_matchable, $param_full_type, $param_type);
    my($param_name, $param_isptr, $param_isarray, $param_array_len);

    foreach $syscall_param_string (@$syscall_params_strings_ref) {
	if (($syscall_param_string =~ /^\s*(matchable|unmatchable)\s+([^\s]+)\s+([^\s]+)\s*$/)
	|| ($syscall_param_string =~ /^\s*()(void)/)) {
	    $param_matchable = $1;
	    $param_full_type = $2;
	    $param_name = $3;
	    $param_isptr = 0;
	    $param_isarray = 0;
	    $param_array_len = -1;
	    if ($param_full_type =~ /^ptr_(.*)$/) {
		$param_isptr = 1;
		$param_type = $1;
	    }
	    elsif ($param_full_type =~ /^array\[([0-9]*)\]_(.*)$/) {
		$param_isarray = 1;
		$param_array_len = $1;
		$param_type = $2;
		if ($param_array_len eq '') {
		    $param_array_len = -1;
		}
	    }
	    else {
		$param_type = $param_full_type;
	    }
	    push(@$syscall_param_matchable_ref, $param_matchable);
	    push(@$syscall_param_types_ref, $param_type);
	    push(@$syscall_param_names_ref, $param_name);
	    push(@$syscall_param_isptr_ref, $param_isptr);
	    push(@$syscall_param_isarray_ref, $param_isarray);
	    push(@$syscall_param_array_len_ref, $param_array_len);
	    # avoid passing unknown types to the logging function by
	    # not encoding them in the param format spec
	    if ($param_isptr == 1) {
		push(@$syscall_param_fmt_spec_ref, "p");
	    }
	    elsif ($param_isarray == 1) {
		# TODO - actually print the array's contents.
		push(@$syscall_param_fmt_spec_ref, "p");
	    }
	    elsif (defined($types_specifier_map{$param_type})) {
		push(@$syscall_param_fmt_spec_ref,
		     $types_specifier_map{$param_type});
	    }
	    elsif ($param_type =~ /^struct_|^const_struct_/) {
		push(@$syscall_param_fmt_spec_ref, "S");
	    }
	}
	else {
	    die "Invalid parameters syntax in '$data_file' line $data_file_line";
	}
    }
}

# handle struct-in-struct.
sub expand_struct_in_struct {
    my($data_file, $data_file_line,
       $pre_struct_param_matchable_ref,
       $pre_struct_param_types_ref,
       $pre_struct_param_names_ref,
       $pre_struct_param_isptr_ref,
       $pre_struct_param_isarray_ref,
       $pre_struct_param_array_len_ref,
       $pre_struct_param_fmt_spec_ref,
       $struct_param_matchable_ref,
       $struct_param_types_ref,
       $struct_param_names_ref,
       $struct_param_isptr_ref,
       $struct_param_isarray_ref,
       $struct_param_array_len_ref,
       $struct_param_fmt_spec_ref) = @_;

    @$struct_param_matchable_ref = ();
    @$struct_param_types_ref = ();
    @$struct_param_names_ref = ();
    @$struct_param_isptr_ref = ();
    @$struct_param_isarray_ref = ();
    @$struct_param_array_len_ref = ();
    @$struct_param_fmt_spec_ref = ();

    my($pre_num_struct_fields) = $#{@$pre_struct_param_names_ref} + 1;

    my($i);
    for ($i=0; $i < $pre_num_struct_fields; $i++) {
	if (@$pre_struct_param_matchable_ref->[$i] eq 'matchable') {
	    if (@$pre_struct_param_types_ref->[$i] =~ /^(const_)?struct_(.*)$/) {
	        my($param_struct_type) = $2;
    		my($param_struct_params) = $structs_fields{$param_struct_type};
		my(@param_struct_params_strings) =
			split(',', $param_struct_params);
		my(@param_struct_param_matchable) = ();
		my(@param_struct_param_types) = ();
		my(@param_struct_param_names) = ();
		my(@param_struct_param_isptr) = ();
		my(@param_struct_param_isarray) = ();
		my(@param_struct_param_array_len) = ();
		my(@param_struct_param_fmt_spec) = ();
    		&break_down_params_list($data_file, $data_file_line,
			    		\@param_struct_params_strings,
			    		\@param_struct_param_matchable,
			    		\@param_struct_param_types,
			    		\@param_struct_param_names,
			    		\@param_struct_param_isptr,
			    		\@param_struct_param_isarray,
			    		\@param_struct_param_array_len,
			    		\@param_struct_param_fmt_spec);

	        push(@$struct_param_matchable_ref, @param_struct_param_matchable);
	        push(@$struct_param_types_ref, @param_struct_param_types);
		my($param_name);
		foreach $param_name (@param_struct_param_names) {
	            push(@$struct_param_names_ref,
			 @$pre_struct_param_names_ref->[$i] . '.' . $param_name);
		}
	        push(@$struct_param_isptr_ref, @param_struct_param_isptr);
	        push(@$struct_param_isarray_ref, @param_struct_param_isarray);
	        push(@$struct_param_array_len_ref, @param_struct_param_array_len);
	        push(@$struct_param_fmt_spec_ref, @param_struct_param_fmt_spec);
	    }
	    else {
	        push(@$struct_param_matchable_ref, $pre_struct_param_matchable_ref->[$i]);
	        push(@$struct_param_types_ref, @$pre_struct_param_types_ref->[$i]);
	        push(@$struct_param_names_ref, @$pre_struct_param_names_ref->[$i]);
	        push(@$struct_param_isptr_ref, @$pre_struct_param_isptr_ref->[$i]);
	        push(@$struct_param_isarray_ref, @$pre_struct_param_isarray_ref->[$i]);
	        push(@$struct_param_array_len_ref, @$pre_struct_param_array_len_ref->[$i]);
	        push(@$struct_param_fmt_spec_ref, @$pre_struct_param_fmt_spec_ref->[$i]);
	    }
	}
	else {
	    push(@$struct_param_matchable_ref, $pre_struct_param_matchable_ref->[$i]);
	    push(@$struct_param_types_ref, @$pre_struct_param_types_ref->[$i]);
	    push(@$struct_param_names_ref, @$pre_struct_param_names_ref->[$i]);
	    push(@$struct_param_isptr_ref, @$pre_struct_param_isptr_ref->[$i]);
	    push(@$struct_param_isarray_ref, @$pre_struct_param_isarray_ref->[$i]);
	    push(@$struct_param_array_len_ref, @$pre_struct_param_array_len_ref->[$i]);
	    push(@$struct_param_fmt_spec_ref, @$pre_struct_param_fmt_spec_ref->[$i]);
	}
    }
}

sub prepare_struct_logging_params {
    my($data_file, $data_file_line, $struct_param, $struct_params_strings_ref) = @_;

    my($struct_logging_params) = "";
    my($struct_param_string);
    my($param_matchable, $param_full_type, $param_type);
    my($param_name, $param_isptr, $param_isarray, $param_array_len);
    my($type_specifier, $dummy_data_specifier);

    foreach $struct_param_string (@$struct_params_strings_ref) {
	if ($struct_param_string =~ /^\s*(matchable|unmatchable)\s+([^\s]+)\s+([^\s]+)\s*$/) {
	    $param_matchable = $1;
	    $param_full_type = $2;
	    $param_name = $3;
	    $param_isptr = 0;
	    $param_isarray = 0;
	    $param_array_len = -1;
	    if ($param_full_type =~ /^ptr_(.*)$/) {
		$param_isptr = 1;
		$param_type = $1;
	    }
	    elsif ($param_full_type =~ /^array\[([0-9]*)\]_(.*)$/) {
		$param_isarray = 1;
		$param_array_len = $1;
		$param_type = $2;
		if ($param_array_len eq '') {
		    $param_array_len = -1;
		}
	    }
	    else {
		$param_type = $param_full_type;
	    }

	    # avoid passing unknown types to the logging function by
	    # not encoding them in the param format spec
	    if (defined($types_specifier_map{$param_type})) {
		$type_specifier = $types_specifier_map{$param_type};
		$dummy_data_specifier =
		    $basic_types_dummy_data_specifier_map{$type_specifier};
    		$struct_logging_params .=
			"(${struct_param} ? "
			. "kern_${struct_param}.${param_name} : "
			. "${dummy_data_specifier}), ";
	    }
	}
	else {
	    die "Invalid struct parameters syntax in '$data_file' line $data_file_line";
	}
    }
    # remove trailing ','.
    $struct_logging_params =~ s/\s*,\s*$//;

    #print "prepare_struct_logging_params: struct_param '$struct_param'\n";
    #print "    struct_logging_params '$struct_logging_params'\n\n";

    return $struct_logging_params;
}

sub emit_structs_data_array() {
    my($structs_fields_ref, $structs_types_indexes_ref) = @_;
    my($struct_type, $struct_type_index);
    my(@struct_types) = keys %$structs_fields_ref;
    my($num_types) = $#struct_types + 1;
    my($line, $line_copy);
    my($data_init) = "";
    my($i) = 0;

    # prepare code to init each entry in the 'struct_type_data' array.
    foreach $struct_type (@struct_types) {
	$struct_type_index = $$structs_types_indexes_ref{$struct_type};
        $data_init .= "\toffsets_vec = sct_make_struct_offsets_vec_${struct_type}(&num_fields, &struct_size);\n";
        $data_init .= "\ttypes_vec = sct_make_struct_types_vec_${struct_type}(&num_fields);\n";
        $data_init .= "\tstruct_type_data[$struct_type_index].offsets_vec = offsets_vec;\n";
        $data_init .= "\tstruct_type_data[$struct_type_index].types_vec = types_vec;\n";
        $data_init .= "\tstruct_type_data[$struct_type_index].num_fields = num_fields;\n";
        $data_init .= "\tstruct_type_data[$struct_type_index].struct_size = struct_size;\n";
        $data_init .= "\tstruct_type_data[$struct_type_index].logger = sct_log_struct_fields_${struct_type};\n";
	$i++;
    }

    foreach $line (@syscall_tracker_struct_data_init_tmpl) {
	# avoid modifying the array - thus, work on a copy of that line.
	$line_copy = $line;
	$line_copy =~ s/%%STRUCT_TYPES_NUM%%/$num_types/g;
	$line_copy =~ s/%%STRUCT_DATA_INIT%%/$data_init/g;
        print TRACKER_AUTOGEN "$line_copy";
    }
    print TRACKER_AUTOGEN "\n";
}

sub emit_stubs_arrays {
    my($num_syscalls, $syscall_ids_ref) = @_;
    my ($syscall_id, $i);

    print("\n", "number syscall ids - ", $#$syscall_ids_ref, "\n");
    print HIJACK_AUTOGEN "/* pointers to stub functions - to be used */\n";
    print HIJACK_AUTOGEN "/* when hijacking system calls.            */\n";
    print HIJACK_AUTOGEN "func_ptr stub_functions[] = {\n";
    print TRACKER_AUTOGEN "/* pointers to 'before' stub functions - to be used */\n";
    print TRACKER_AUTOGEN "/* when hijacking system calls.                     */\n";
    print TRACKER_AUTOGEN "func_ptr before_track_functions[] = {\n";
    for (0..$num_syscalls) {
	$syscall_id = $_;
	# check if we have a stub function for this syscall.
	for ($i=0; $i <= $#$syscall_ids_ref; $i++) {
		if ($syscall_ids_ref->[$i] == $syscall_id) {
			last;
		}
	}
	if ($i <= $#$syscall_ids_ref) {
		print HIJACK_AUTOGEN "\t(func_ptr)stub_syscall_${_},\n";
		print TRACKER_AUTOGEN "\t(func_ptr)track_before_syscall_${_},\n";
	}
	else {
		print HIJACK_AUTOGEN "\t(func_ptr)NULL,\n";
		print TRACKER_AUTOGEN "\t(func_ptr)NULL,\n";
	}
    }
    print HIJACK_AUTOGEN "};\n";
    print TRACKER_AUTOGEN "};\n";

    print TRACKER_AUTOGEN "\n";
    print TRACKER_AUTOGEN "/* pointers to 'after' stub functions - to be used */\n";
    print TRACKER_AUTOGEN "/* when hijacking system calls.                     */\n";
    print TRACKER_AUTOGEN "func_ptr after_track_functions[] = {\n";
    for (0..$num_syscalls) {
	$syscall_id = $_;
	# check if we have a stub function for this syscall.
	for ($i=0; $i <= $#$syscall_ids_ref; $i++) {
		if ($syscall_ids_ref->[$i] == $syscall_id) {
			last;
		}
	}
	if ($i <= $#$syscall_ids_ref) {
		print TRACKER_AUTOGEN "\t(func_ptr)track_after_syscall_${_},\n";
	}
	else {
		print TRACKER_AUTOGEN "\t(func_ptr)NULL,\n";
	}
    }
    print TRACKER_AUTOGEN "};\n";
}
