/* 
 * $Id: sct_rules_private.h,v 1.6 2002/02/02 15:56:36 mulix Exp $
 */

#ifndef SCT_RULES_PRIVATE_H_
#define SCT_RULES_PRIVATE_H_ 

#include <linux/stddef.h>
#include <linux/sys.h>

#include "syscall_hijack_public.h"
#include "rule_engine_public.h"

/* the tracker mutex. all operations involving filter_vars or meta data */
/* rule lists, etc) must take the lock.                                 */
extern struct semaphore tracker_sem;

/* we only support one mux function now, socketcall.               */
/* you need to update this section if you add more muxed syscalls. */
#define NR_MUX_SYSCALLS 1
#define MUX_NO_SYSCALL -1 /* no such muxed call */
#define MUX_SOCKETCALL 0  /* for socketcall */

/* translate a syscall_id for a mux syscall into an index to */
/* the mux_data array.                                       */
static inline int mux_data_index(int syscall_id)
{
	switch (syscall_id) {
		/* for socketcall */
	case (102): return MUX_SOCKETCALL;
	
        /* other mux syscalls would go here */
	
	/* for anything else */
	default: return MUX_NO_SYSCALL;
	}
}

struct mux_data_t {
	func_ptr mux_tracker;	    /* tracker for the syscall itself.      */
	int num_muxed;		    /* number of muxed functions.           */
	func_ptr* muxed_functions;  /* trackers for the muxed functions.    */
	func_ptr* demux_functions;  /* functions allowing re-calling the    */
	                            /* orig syscall from the function       */
	                            /* trackers.                            */
	sct_rule_list_t** before_rules; /* rules list, one per muxed function.  */
	sct_rule_list_t** after_rules;  /* rules list, one per muxed function.  */
};

extern struct mux_data_t mux_data[NR_MUX_SYSCALLS];

/* extract native syscall id from a syscall id. */
#define	SYSCALL_PART_OF_SYSCALL_ID(syscall_id)	((syscall_id) & 0xffff)
/* extract multiplexed function id from a syscall id. */
#define	FUNC_PART_OF_SYSCALL_ID(syscall_id)	(((syscall_id) >> 16) & 0xffff)

/* type of info stored for a tracked syscall. */
struct syscall_track_data {
	sct_rule_list_t* before_rules;  /* list of 'before' rules for syscall. */
	sct_rule_list_t* after_rules;   /* list of 'after' rules for syscall.  */
};

/* track data for all syscalls. */
extern struct syscall_track_data track_data[NR_syscalls];

/* lives in syscall_tracker_autogen.c */
extern func_ptr before_track_functions[];
extern func_ptr after_track_functions[];

#endif /* SCT_RULES_PRIVATE_H_ */
